import numpy as np
import os
import matplotlib.pyplot as plt
from read_cifar import read_cifar, split_dataset


# Fonction d'activation sigmoid
def sigmoid(x):
    return 1 / (1 + np.exp(-x))

# Dérivée de la fonction d'activation sigmoid
def sigmoid_derivative(x):
    return x * (1 - x)
# Fonction d'erreur quadratique moyenne

def mean_squared_error(predictions, targets):
    return np.mean(np.square(predictions - targets))

# Conversion des étiquettes en one-hot encoding
def one_hot(labels, num_classes):
    one_hot_labels = np.zeros((len(labels), num_classes))
    one_hot_labels[np.arange(len(labels)), labels] = 1
    return one_hot_labels

# Fonction pour une itération d'apprentissage avec la perte de mean squared error (MSE)
def learn_once_mse(w1, b1, w2, b2, data, targets, learning_rate):
    a0 = data
    z1 = np.dot(a0, w1) + b1
    a1 = sigmoid(z1)
    z2 = np.dot(a1, w2) + b2
    a2 = sigmoid(z2)

    dC_da2 = a2 - targets
    dC_dz2 = dC_da2 * sigmoid_derivative(a2)
    dC_da1 = np.dot(dC_dz2, w2.T)
    dC_dz1 = dC_da1 * sigmoid_derivative(a1)

    dC_dw2 = np.dot(a1.T, dC_dz2)
    dC_db2 = np.sum(dC_dz2, axis=0)
    dC_dw1 = np.dot(a0.T, dC_dz1)
    dC_db1 = np.sum(dC_dz1, axis=0)

    # moyenne des gradients
    w1 -= learning_rate * dC_dw1 / len(data)
    b1 -= learning_rate * dC_db1 / len(data)
    w2 -= learning_rate * dC_dw2 / len(data)
    b2 -= learning_rate * dC_db2 / len(data)

    loss = mean_squared_error(a2, targets)

    return w1, b1, w2, b2, loss

# Initialisation des poids 
def initialize_weights(input_size, hidden_size, output_size):
    
    w1 = np.random.randn(input_size, hidden_size) / np.sqrt(input_size)
    b1 = np.zeros((1, hidden_size))
    w2 = np.random.randn(hidden_size, output_size) / np.sqrt(hidden_size)
    b2 = np.zeros((1, output_size))

    return w1, b1, w2, b2

# Ajout de la fonction de prédiction
def predict_mlp(w1, b1, w2, b2, data):
    a0 = data
    z1 = np.dot(a0, w1) + b1
    a1 = sigmoid(z1)
    z2 = np.dot(a1, w2) + b2
    a2 = sigmoid(z2)
    return a2


# Fonction principale pour le training du MLP

def run_mlp_training(data_train, labels_train, data_test, labels_test, d_h, learning_rate, num_epochs, results_dir='results'):
    # Create results directory if it does not exist
    if not os.path.exists(results_dir):
        os.makedirs(results_dir)

    N_train = len(data_train)
    N_test = len(data_test)
    d_in = data_train.shape[1]
    d_out = len(np.unique(labels_train))

    w1, b1, w2, b2 = initialize_weights(d_in, d_h, d_out)

    train_accuracies = []
    losses = []

    for epoch in range(num_epochs):
        w1, b1, w2, b2, loss = learn_once_mse(w1, b1, w2, b2, data_train, one_hot(labels_train, d_out), learning_rate)

        predictions = predict_mlp(w1, b1, w2, b2, data_train)
        accuracy = np.mean(np.argmax(predictions, axis=1) == labels_train)
        train_accuracies.append(accuracy)
        losses.append(loss)
        print(f"Epoch {epoch + 1}/{num_epochs} - Loss: {loss:.4f} - Accuracy: {accuracy:.4f}")

    # Enregistrement du graphe training accuracy
    plt.figure(figsize=(8, 6))
    plt.plot(range(1, num_epochs + 1), train_accuracies)
    plt.xlabel("Epoch")
    plt.ylabel("Training Accuracy")
    plt.title("MLP Training Accuracy Evolution")
    plt.grid(True)

    save_path = os.path.join(results_dir, "mlp_training_accuracy.png")
    plt.savefig(save_path)
    plt.show()


# Avec les données on a :
split_factor = 0.9
folder_path = 'data/cifar-10-batches-py'
data, labels = read_cifar(folder_path)

data_train, data_test, labels_train, labels_test = split_dataset(data, labels, split_factor)
d_h = 64
learning_rate = 0.01
num_epochs = 100

# Pour que ça aille dans le dossier results dans le même dossier que le code
results_directory = 'results'

# Exécution de l'entraînement mlp
w1, b1, w2, b2, train_accuracies, losses = run_mlp_training(data_train, labels_train, data_test, labels_test, d_h, learning_rate, num_epochs, results_directory)
